@@ -2,11 +2,11 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, Workflow, fields
-from trytond.pool import Pool, PoolMeta
-from trytond.pyson import Eval, If, Bool
-
+from trytond.model import Index, ModelSQL, ModelView, Workflow, fields
 from trytond.modules.product import price_digits
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If
+from trytond.transaction import Transaction
 
 from .exceptions import AmendmentValidateError
 
@@ -17,11 +17,11 @@ class Sale(metaclass=PoolMeta):
     amendments = fields.One2Many(
         'sale.amendment', 'sale', "Amendments",
         states={
-            'invisible': ((Eval('state') != 'processing')
-                | ~Eval('amendments')),
+            'invisible': (
+                ~Eval('state').in_(['processing', 'done'])
+                | ((Eval('state') == 'done') & ~Eval('amendments'))),
             'readonly': Eval('state') != 'processing',
-            },
-        depends=['state'])
+            })
 
     @classmethod
     def copy(cls, sales, default=None):
@@ -38,40 +38,44 @@ class Amendment(Workflow, ModelSQL, ModelView):
     __name__ = 'sale.amendment'
 
     sale = fields.Many2One(
-        'sale.sale', "Sale", required=True, select=True,
+        'sale.sale', "Sale", required=True,
         domain=[
             ('state', 'in', ['processing', 'done']),
             ],
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
-            },
-        depends=['state'])
+            })
     date = fields.Date(
         "Date", required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     description = fields.Char(
         "Description",
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     state = fields.Selection([
             ('draft', "Draft"),
             ('validated', "Validated"),
-            ], "State", readonly=True, required=True)
+            ], "State", readonly=True, required=True, sort=False)
     lines = fields.One2Many(
         'sale.amendment.line', 'amendment', "Lines",
         states={
-            'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            'readonly': (
+                (Eval('state') != 'draft')
+                | ~Eval('sale')),
+            })
 
     @classmethod
     def __setup__(cls):
         super(Amendment, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state == 'draft'))
         cls._order = [
             ('date', 'DESC'),
             ('id', 'DESC'),
@@ -100,6 +104,8 @@ class Amendment(Workflow, ModelSQL, ModelView):
     def validate_amendment(cls, amendments):
         pool = Pool()
         Sale = pool.get('sale.sale')
+        transaction = Transaction()
+        context = transaction.context
         sales = set()
         for amendment in amendments:
             sale = amendment.sale
@@ -121,7 +127,9 @@ class Amendment(Workflow, ModelSQL, ModelView):
         Sale.save(sales)
         Sale.store_cache(sales)
         cls._clear_sale(sales)
-        Sale.__queue__.process(sales)
+        with transaction.set_context(
+                queue_batch=context.get('queue_batch', True)):
+            Sale.__queue__.process(sales)
 
     @classmethod
     def _clear_sale(cls, sales):
@@ -190,8 +198,7 @@ class AmendmentLine(ModelSQL, ModelView):
             ], "Action", required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
 
     sale = fields.Function(fields.Many2One(
             'sale.sale', "Sale"), 'on_change_with_sale')
@@ -205,15 +212,13 @@ class AmendmentLine(ModelSQL, ModelView):
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
             'required': Eval('action') == 'line',
-            },
-        depends=['state', 'sale', 'action'])
+            })
 
     payment_term = fields.Many2One(
-        'account.invoice.payment_term', "Payment Term",
+        'account.invoice.payment_term', "Payment Term", ondelete='RESTRICT',
         states={
             'invisible': Eval('action') != 'payment_term',
-            },
-        depends=['action'])
+            })
 
     party = fields.Many2One(
         'party.party', "Party",
@@ -221,38 +226,48 @@ class AmendmentLine(ModelSQL, ModelView):
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'party',
             'required': Eval('action') == 'party',
+            })
+    invoice_party = fields.Many2One(
+        'party.party', "Invoice Party",
+        states={
+            'readonly': Eval('state') != 'draft',
+            'invisible': Eval('action') != 'party',
             },
-        depends=['state', 'action'])
+        search_context={
+            'related_party': Eval('party'),
+            })
     invoice_address = fields.Many2One(
         'party.address', "Invoice Address",
         domain=[
-            ('party', '=', Eval('party')),
+            ('party', '=', If(Eval('invoice_party'),
+                    Eval('invoice_party'), Eval('party'))),
             ],
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'party',
             'required': Eval('action') == 'party',
-            },
-        depends=['state', 'action', 'party'])
+            })
     shipment_party = fields.Many2One(
         'party.party', "Shipment Party",
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'party',
             },
-        depends=['state', 'action'])
+        search_context={
+            'related_party': Eval('party'),
+            })
     shipment_address = fields.Many2One(
         'party.address', "Shipment Address",
-        domain=[
+        domain=['OR',
             ('party', '=', If(Eval('shipment_party'),
                     Eval('shipment_party'), Eval('party'))),
+            ('warehouses', '=', Eval('warehouse')),
             ],
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'party',
             'required': Eval('action') == 'party',
-            },
-        depends=['state', 'action', 'party', 'shipment_party'])
+            })
 
     warehouse = fields.Many2One(
         'stock.location', "Warehouse",
@@ -263,13 +278,13 @@ class AmendmentLine(ModelSQL, ModelView):
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'warehouse',
             'required': Eval('action') == 'warehouse',
-            },
-        depends=['state', 'action'])
+            })
 
     product = fields.Many2One(
         'product.product', "Product",
         domain=[
-            If(Eval('state') == 'draft',
+            If((Eval('state') == 'draft')
+                & ~(Eval('quantity', 0) < 0),
                 ('salable', '=', True),
                 ()),
             If(Eval('product_uom_category'),
@@ -279,17 +294,14 @@ class AmendmentLine(ModelSQL, ModelView):
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
-            },
-        depends=['action', 'state', 'product_uom_category'])
+            })
     quantity = fields.Float(
-        "Quantity",
-        digits=(16, Eval('unit_digits', 2)),
+        "Quantity", digits='unit',
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
             'required': Eval('action') == 'line',
-            },
-        depends=['unit_digits', 'action'])
+            })
     unit = fields.Many2One(
         'product.uom', "Unit", ondelete='RESTRICT',
         domain=[
@@ -301,27 +313,21 @@ class AmendmentLine(ModelSQL, ModelView):
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
             'required': Bool(Eval('product')),
-            },
-        depends=['state', 'product_uom_category', 'action', 'product'])
+            })
     unit_price = fields.Numeric(
         "Unit Price", digits=price_digits,
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
             'required': Eval('action') == 'line',
-            },
-        depends=['state', 'action'])
+            })
     description = fields.Text(
         "Description",
         states={
             'readonly': Eval('state') != 'draft',
             'invisible': Eval('action') != 'line',
-            },
-        depends=['state', 'action'])
+            })
 
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"),
-        'on_change_with_unit_digits')
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', "Product UoM Category"),
         'on_change_with_product_uom_category')
@@ -336,6 +342,7 @@ class AmendmentLine(ModelSQL, ModelView):
         '_parent_amendment.sale',
         '_parent_amendment._parent_sale.payment_term',
         '_parent_amendment._parent_sale.party',
+        '_parent_amendment._parent_sale.invoice_party',
         '_parent_amendment._parent_sale.invoice_address',
         '_parent_amendment._parent_sale.shipment_party',
         '_parent_amendment._parent_sale.shipment_address',
@@ -345,6 +352,7 @@ class AmendmentLine(ModelSQL, ModelView):
             self.payment_term = self.amendment.sale.payment_term
 
             self.party = self.amendment.sale.party
+            self.invoice_party = self.amendment.sale.invoice_party
             self.invoice_address = self.amendment.sale.invoice_address
             self.shipment_party = self.amendment.sale.shipment_party
             self.shipment_address = self.amendment.sale.shipment_address
@@ -378,8 +386,7 @@ class AmendmentLine(ModelSQL, ModelView):
         'amendment',
         '_parent_amendment.sale')
     def on_change_with_sale(self, name=None):
-        if self.amendment and self.amendment.sale:
-            return self.amendment.sale.id
+        return self.amendment.sale if self.amendment else None
 
     @fields.depends('line')
     def on_change_line(self):
@@ -393,36 +400,47 @@ class AmendmentLine(ModelSQL, ModelView):
             self.product = self.quantity = self.unit = self.unit_price = None
             self.description = None
 
-    @fields.depends('party', 'shipment_party')
+    @fields.depends('party', 'invoice_party', 'shipment_party', 'warehouse')
     def on_change_party(self):
-        self.invoice_address = None
+        if not self.invoice_party:
+            self.invoice_address = None
         if not self.shipment_party:
             self.shipment_address = None
         if self.party:
-            self.invoice_address = self.party.address_get(type='invoice')
+            if not self.invoice_address:
+                self.invoice_address = self.party.address_get(type='invoice')
             if not self.shipment_party:
-                self.shipment_address = self.party.address_get(type='delivery')
+                with Transaction().set_context(
+                        warehouse=(
+                            self.warehouse.id if self.warehouse else None)):
+                    self.shipment_address = self.party.address_get(
+                        type='delivery')
 
-    @fields.depends('party', 'shipment_party')
-    def on_change_shipment_party(self):
-        if self.shipment_party:
-            self.shipment_address = self.shipment_party.address_get(
-                type='delivery')
+    @fields.depends('party', 'invoice_party')
+    def on_change_invoice_party(self):
+        if self.invoice_party:
+            self.invoice_address = self.invoice_party.address_get(
+                type='invoice')
         elif self.party:
-            self.shipment_address = self.party.address_get(type='delivery')
+            self.invoice_address = self.party.address_get(type='invoice')
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if self.unit:
-            return self.unit.digits
+    @fields.depends('party', 'shipment_party', 'warehouse')
+    def on_change_shipment_party(self):
+        with Transaction().set_context(
+                warehouse=self.warehouse.id if self.warehouse else None):
+            if self.shipment_party:
+                self.shipment_address = self.shipment_party.address_get(
+                    type='delivery')
+            elif self.party:
+                self.shipment_address = self.party.address_get(type='delivery')
 
     @fields.depends('line')
     def on_change_with_product_uom_category(self, name=None):
         if self.line:
             if self.line.product_uom_category:
-                return self.line.product_uom_category.id
+                return self.line.product_uom_category
             elif self.line.unit:
-                return self.line.unit.category.id
+                return self.line.unit.category
 
     def apply(self, sale):
         assert self.sale == sale
@@ -444,6 +462,7 @@ class AmendmentLine(ModelSQL, ModelView):
 
     def _apply_party(self, sale, sale_line):
         sale.party = self.party
+        sale.invoice_party = self.invoice_party
         sale.invoice_address = self.invoice_address
         sale.shipment_party = self.shipment_party
         sale.shipment_address = self.shipment_address
